﻿using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{
    public class AudioFileManager : MonoBehaviour
    {
        private bool _isLoading;

        public class SceneFilesUpdatedEvent : UnityEvent { }
        public readonly SceneFilesUpdatedEvent OnSceneFilesUpdated = new SceneFilesUpdatedEvent();

        public readonly List<NamedAudioClip> AddTheseFilesToActiveCollection = new List<NamedAudioClip>();

        public AudioFileManager()
        {
            _isLoading = false;
        }

        public void OpenImportFolderDialog(bool addFilesToActiveCollection = false)
        {
            SuperController.singleton.GetDirectoryPathDialog((path) => ImportFolder(path, null, addFilesToActiveCollection), "Custom/Sounds");
        }

        public void OpenImportFileDialog(bool addFilesToActiveCollection = false)
        {
            SuperController.singleton.GetMediaPathDialog((path) => ImportFile(path, addFilesToActiveCollection), "mp3|ogg|wav", "Custom/Sounds");
        }

        public void ImportFolder(string path, UnityAction callback = null, bool addFilesToActiveCollection = false)
        {
            if (_isLoading) return;
            try
            {
                _isLoading = true;
                if (path == "") return;
                var files = SuperController.singleton.GetFilesAtPath(path).ToList();
                files.ForEach((string fileName) =>
                {
                    var isValid = !fileName.Contains(".json", StringComparison.OrdinalIgnoreCase) &&
                                  (fileName.Contains(".mp3", StringComparison.OrdinalIgnoreCase)
                                   || fileName.Contains(".wav", StringComparison.OrdinalIgnoreCase)
                                   || fileName.Contains(".ogg", StringComparison.OrdinalIgnoreCase));
                    if (!isValid) return;

                    var loadedClip = Load(fileName);
                    if (addFilesToActiveCollection && loadedClip != null)
                    {
                        AddTheseFilesToActiveCollection.Add(loadedClip);
                    }
                });
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioFileManager)}.{nameof(ImportFolder)}: {e}");
            }
            finally
            {
                _isLoading = false;
                OnSceneFilesUpdated.Invoke();
                callback?.Invoke();
            }
        }

        private void ImportFile(string path, bool addFileToActiveCollection = false)
        {
            var loadedClip = Load(path);
            if (addFileToActiveCollection && loadedClip != null)
            {
                AddTheseFilesToActiveCollection.Add(loadedClip);
            }
            OnSceneFilesUpdated.Invoke();
        }

        public static NamedAudioClip Load(string path)
        {
            var normalizedPath = SuperController.singleton.NormalizeLoadPath(path);
            var existing = URLAudioClipManager.singleton.GetClip(normalizedPath);
            return existing != null ? null : URLAudioClipManager.singleton.QueueClip(normalizedPath);
        }
    }
}
