using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using Image = UnityEngine.UI.Image;

namespace AudioMate
{
    public class HorizontalButtonGroup : MonoBehaviour
    {
        protected GameObject Container;
        public HorizontalLayoutGroup Layout;

        public RectOffset Padding
        {
            set { Layout.padding = value; }
        }

        public float Spacing
        {
            set { Layout.spacing = value; }
        }

        protected HorizontalButtonGroup()
        {
            Init();
        }

        public void Init()
        {
            Container = new GameObject($"HorizontalGroupContainer{Tools.GenerateID()}");
            Container.transform.SetParent(transform, false);
            var rectTransform = Container.AddComponent<RectTransform>();
            rectTransform.StretchParent();
            Layout = Container.AddComponent<HorizontalLayoutGroup>();
            Layout.spacing = 10f;
            Layout.childForceExpandWidth = false;
            Layout.childControlWidth = true;
            Layout.childForceExpandHeight = true;
            Layout.childControlHeight = true;
            Layout.childAlignment = TextAnchor.MiddleCenter;
        }

        public virtual UIDynamicButton CreateButton(string label, int style = Styles.Button, UnityAction call = null,
            bool addOutline = false, float preferredWidth = 100f, float flexibleWidth = 1, int fontSize = 30, float preferredHeight = 40f, float flexibleHeight = 40f)
        {
            var instance = Instantiate(UIFactory.ButtonPrefab, Container.transform, false);
            var btn = instance.GetComponent<UIDynamicButton>();
            btn.label = label;

            btn.textColor = Styles.Text(style);
            btn.buttonColor = Styles.Bg(style);
            btn.buttonText.fontSize = fontSize;
            if (addOutline)
            {
                var outline = btn.button.gameObject.AddComponent<Outline>();
                outline.effectColor = Styles.Text(style);
                outline.effectDistance = new Vector2(3f, 3f);
            }
            var layout = instance.GetComponent<LayoutElement>();
            layout.preferredWidth = preferredWidth;
            layout.preferredHeight = preferredHeight;
            layout.flexibleWidth = flexibleWidth;
            layout.flexibleHeight = flexibleHeight;
            layout.minWidth = 20f;

            if (call != null) btn.button.onClick.AddListener(call);

            return btn;
        }

        public virtual Text CreateLabel(string text, int style = Styles.Button, bool invertStyle = false, int size = 28, bool bold = false, TextAnchor alignment = TextAnchor.MiddleLeft, float minWidth = 0f)
        {
            var go = new GameObject("Label");
            go.transform.SetParent(Container.transform, false);
            var layout = go.AddComponent<LayoutElement>();
            layout.flexibleWidth = 1f;
            layout.flexibleHeight = 1f;
            layout.preferredWidth = 1f;
            layout.preferredHeight = 40f;
            layout.minWidth = minWidth;

            var label = go.gameObject.AddComponent<Text>();
            label.text = text;
            label.color = invertStyle ? Styles.Bg(style) : Styles.Text(style);
            label.font = UIFactory.BaseFont;
            label.fontSize = size;
            label.fontStyle = bold ? FontStyle.Bold : FontStyle.Normal;
            label.verticalOverflow = VerticalWrapMode.Overflow;
            label.alignment = alignment;

            return label;
        }

        private bool _hasBackground;

        public virtual void SetBg(Color color)
        {
            Image image;
            if (_hasBackground)
            {
                image = Container.GetComponent<Image>();
            }
            else
            {
                image = Container.AddComponent<Image>();
                _hasBackground = true;
            }
            image.rectTransform.StretchParent();
            image.fillAmount = 1f;
            image.fillCenter = true;
            image.fillClockwise = true;
            image.fillOrigin = 0;
            image.preserveAspect = false;
            image.color = color;
        }

        public virtual void Destroy()
        {
            Destroy(Container);
        }

    }
}
