﻿using System.Collections;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;

namespace AudioMate
{
    public class MouseOverDetector : MonoBehaviour
    {
        private RectTransform _rectTransform;
        private Vector2 _leftBottom;
        private Vector2 _rightTop;
        private bool _init;
        private bool _hasEnteredEventFired;
        private bool _hasExitedEventFired;

        public bool isOver;

        public class MouseOverStateChangedEvent : UnityEvent {};
        public readonly MouseOverStateChangedEvent OnMouseEntered = new MouseOverStateChangedEvent();
        public readonly MouseOverStateChangedEvent OnMouseExited = new MouseOverStateChangedEvent();
        public void Start()
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);
            _rectTransform = go.AddComponent<RectTransform>();
            _rectTransform.StretchParent();
            go.AddComponent<LayoutElement>();
            _init = false;
            StartCoroutine(InitPositionData());
        }

        public void Refresh()
        {
            _init = false;
            _hasEnteredEventFired = false;
            _hasExitedEventFired = false;
            StartCoroutine(InitPositionData());
        }

        private void FixedUpdate()
        {
            if (!_init)
            {
                isOver = false;
                return;
            }
            Vector3 mousePos = Input.mousePosition;


            if (mousePos.x >= _leftBottom.x
                && mousePos.x <= _rightTop.x
                && mousePos.y >= _leftBottom.y
                && mousePos.y <= _rightTop.y)
            {
                if (!_hasEnteredEventFired)
                {
                    OnMouseEntered.Invoke();
                    _hasEnteredEventFired = true;
                    _hasExitedEventFired = false;

                }
                isOver = true;
            }
            else
            {
                if (!_hasExitedEventFired)
                {
                    OnMouseExited.Invoke();
                    _hasExitedEventFired = true;
                    _hasEnteredEventFired = false;
                }
                isOver = false;
            }
        }

        private IEnumerator InitPositionData()
        {
            yield return null;
            Vector3[] corners = new Vector3[4];
            _rectTransform.GetWorldCorners(corners);
            _leftBottom = RectTransformUtility.WorldToScreenPoint(Camera.main, corners[0]);
            _rightTop = RectTransformUtility.WorldToScreenPoint(Camera.main, corners[2]);
            isOver = false;
            _init = true;
        }
    }
}
