﻿using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{
    public abstract class Screen : MonoBehaviour
    {
        public class ScreenChangeRequestedEvent : UnityEvent<string> { }

        public static readonly ScreenDefinition Definition = new ScreenDefinition(
            "",
            "",
            typeof(Screen));

        public string GetID => Definition.ID;
        public string GetTitle => Definition.Title;

        public readonly ScreenChangeRequestedEvent OnScreenChangeRequested = new ScreenChangeRequestedEvent();

        protected AudioMateController C;
        protected UIFactory UI;
        protected CollectionManager Collections => C != null ? C.collections : null;
        protected Collection ActiveCollection => Collections != null ? Collections.ActiveCollection : null;

        private bool _disposing;

        public virtual void Init(AudioMateController controller)
        {
            C = controller;
            UI = gameObject.AddComponent<UIFactory>();
            UI.c = controller;
            C.collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);
            C.collections.OnActiveCollectionUpdated.AddListener(OnActiveCollectionUpdated);
            C.collections.OnActiveCollectionNameChanged.AddListener(OnActiveCollectionNameChanged);
            _Init();
        }

        public virtual void _Init()
        {
            // ...
        }

        protected virtual void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
            // ...
        }

        protected virtual void OnActiveCollectionUpdated()
        {
            // ...
        }

        protected virtual void OnActiveCollectionNameChanged(CollectionManager.ActiveCollectionNameChangedEventArgs args)
        {
            // ...
        }

        public virtual void OnDestroy()
        {
            if (_disposing) return;
            _disposing = true;
            Unbind();
        }

        public virtual void Unbind()
        {
            C.collections.OnActiveCollectionSelected.RemoveListener(OnActiveCollectionSelected);
            C.collections.OnActiveCollectionUpdated.RemoveListener(OnActiveCollectionUpdated);
            C.collections.OnActiveCollectionNameChanged.RemoveListener(OnActiveCollectionNameChanged);
        }

        /* public virtual void Refresh()
        {
            Unbind();
            Destroy(UI);
            Init(C);
        } */

        protected virtual void Log(string message)
        {
            if ((Object) C == (Object) null) return;
            C.Log($" [{nameof(Screen)}]: {message}");
        }
    }
}
