﻿using UnityEngine;
using UnityEngine.UI;

namespace AudioMate
{
    public class ScreenSettings : Screen
    {
        public new static readonly ScreenDefinition Definition = new ScreenDefinition(
            "Settings",
            "Global",
            typeof(ScreenSettings));

        private JSONStorableBool _toggleDisplayHints;
        private JSONStorableBool _toggleVolumeControlEnabled;
        private JSONStorableBool _togglePitchControlEnabled;
        private JSONStorableBool _toggleAdvancedUI;
        private JSONStorableBool _toggleAddImportedFilesToActiveCollection;
        private JSONStorableFloat _floatFirstMultiselectButtonAmount;
        private JSONStorableFloat _floatSecondMultiselectButtonAmount;

        public override void Init(AudioMateController controller)
        {
            base.Init(controller);
            InitSettingsUI();
        }

        private void InitSettingsUI()
        {
            UI.CreateHeader("Global Settings", 2, 20f);

            _toggleDisplayHints = new JSONStorableBool(
                "Display Hints",
                C.DisplayHintsJSON.val,
                val => C.DisplayHintsJSON.val = val);
            UI.CreateToggle(_toggleDisplayHints);


            UI.CreateHeader("Clip Library Settings");

            _toggleAdvancedUI = new JSONStorableBool(
                "Show Remove Button For Clips",
                C.AdvancedUIEnabledJSON.val,
                val => C.AdvancedUIEnabledJSON.val = val);
            UI.CreateToggle(_toggleAdvancedUI);

            _toggleAddImportedFilesToActiveCollection = new JSONStorableBool(
                "Add Imported Files To Active Collection",
                C.AddImportedFilesToActiveCollectionJSON.val,
                val => C.AddImportedFilesToActiveCollectionJSON.val = val);
            UI.CreateToggle(_toggleAddImportedFilesToActiveCollection);

            _floatFirstMultiselectButtonAmount = new JSONStorableFloat(
                "First Multiselect Amount",
                C.FirstMultiselectButtonAmountJSON.defaultVal,
                val => C.FirstMultiselectButtonAmountJSON.val = val,
                2,
                100);
            UI.CreateSlider(_floatFirstMultiselectButtonAmount, 1f, false, "F0");
            _floatFirstMultiselectButtonAmount.valNoCallback = C.FirstMultiselectButtonAmountJSON.val;

            _floatSecondMultiselectButtonAmount = new JSONStorableFloat(
                "Second Multiselect Amount",
                C.SecondMultiselectButtonAmountJSON.defaultVal,
                val => C.SecondMultiselectButtonAmountJSON.val = val,
                2,
                100);
            UI.CreateSlider(_floatSecondMultiselectButtonAmount, 1f, false, "F0");
            _floatSecondMultiselectButtonAmount.valNoCallback = C.SecondMultiselectButtonAmountJSON.val;


            UI.CreateHeader("Collection Settings");

            _toggleVolumeControlEnabled = new JSONStorableBool(
                "Enable Volume Control",
                C.VolumeControlEnabledJSON.val,
                val => C.VolumeControlEnabledJSON.val = val);
            UI.CreateToggle(_toggleVolumeControlEnabled);

            _togglePitchControlEnabled = new JSONStorableBool(
                "Enable Pitch Control",
                C.PitchControlEnabledJSON.val,
                val => C.PitchControlEnabledJSON.val = val);
            UI.CreateToggle(_togglePitchControlEnabled);
        }
    }
}
