using System;
using System.Collections;
using System.Collections.Generic;
using AssetBundles;
using UnityEngine;
using UnityEngine.UI;
using Image = UnityEngine.UI.Image;
using Object = UnityEngine.Object;

namespace AudioMate
{
    /// <summary>
    /// Base code shamelessly ripped from the incredible Timeline plugin (https://github.com/acidbubbles/vam-timeline)
    /// Shout outs to Acidbubbles!
    /// </summary>
    public class UIFactory : MonoBehaviour
    {
        private static RectTransform _scrollbarPrefab;
        private static RectTransform _triggerActionsPrefab;
        private static RectTransform _triggerActionMiniPrefab;
        private static RectTransform _triggerActionDiscretePrefab;
        private static RectTransform _triggerActionTransitionPrefab;
        public static bool hasLoadedAssets;
        public RectTransform TriggerActionsPrefab
        {
            get { return _triggerActionsPrefab; }
        }

        public RectTransform TriggerActionMiniPrefab
        {
            get { return _triggerActionMiniPrefab; }
        }

        public RectTransform TriggerActionDiscretePrefab
        {
            get { return _triggerActionDiscretePrefab; }
        }

        public RectTransform TriggerActionTransitionPrefab
        {
            get { return _triggerActionTransitionPrefab; }
        }

        public static RectTransform ButtonPrefab;
        public static readonly Font BaseFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
        private GameObject _currentContainer;

        public static IEnumerator LoadUIAssets()
        {
            foreach (var x in LoadUIAsset(
                         "z_ui2",
                         "DynamicTextField",
                         prefab => _scrollbarPrefab = prefab.GetComponentInChildren<ScrollRect>().verticalScrollbar.gameObject.GetComponent<RectTransform>())
                     ) yield return x;
            foreach (var x in LoadUIAsset(
                         "z_ui2",
                         "DynamicButton",
                         prefab => ButtonPrefab = prefab)) yield return x;
            foreach (var x in LoadUIAsset(
                         "z_ui2",
                         "TriggerActionsPanel",
                         prefab => _triggerActionsPrefab = prefab)) yield return x;
            foreach (var x in LoadUIAsset(
                         "z_ui2",
                         "TriggerActionMiniPanel",
                         prefab => _triggerActionMiniPrefab = prefab)) yield return x;
            foreach (var x in LoadUIAsset(
                         "z_ui2",
                         "TriggerActionDiscretePanel",
                         prefab => _triggerActionDiscretePrefab = prefab)) yield return x;
            foreach (var x in LoadUIAsset(
                         "z_ui2",
                         "TriggerActionTransitionPanel",
                         prefab => _triggerActionTransitionPrefab = prefab)) yield return x;
            hasLoadedAssets = true;
        }

        private static IEnumerable LoadUIAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            var request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null) throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            var go = request.GetAsset<GameObject>();
            if (go == null) throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            var prefab = go.GetComponent<RectTransform>();
            if (prefab == null) throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }

        public static ScrollRect CreateScrollRect(GameObject gameObject, RectOffset padding = null)
        {
            var scrollView = CreateScrollView(gameObject);
            var viewport = CreateViewport(scrollView);
            var content = CreateContent(viewport, padding);
            var scrollbar = CreateScrollbar(scrollView);
            var scrollRect = scrollView.GetComponent<ScrollRect>();
            scrollRect.viewport = viewport.GetComponent<RectTransform>();
            scrollRect.content = content.GetComponent<RectTransform>();
            scrollRect.verticalScrollbar = scrollbar;
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
            return scrollRect; //content.GetComponent<RectTransform>();
        }

        private static GameObject CreateScrollView(GameObject parent)
        {
            var go = new GameObject("Scroll View");
            go.transform.SetParent(parent.transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchParent();

            var scroll = go.AddComponent<ScrollRect>();
            scroll.horizontal = false;

            return go;
        }

        private static Scrollbar CreateScrollbar(GameObject scrollView)
        {
            var vs = Instantiate(_scrollbarPrefab, scrollView.transform, false);
            return vs.GetComponent<Scrollbar>();
        }

        private static GameObject CreateViewport(GameObject scrollView)
        {
            var go = new GameObject("Viewport");
            go.transform.SetParent(scrollView.transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchParent();
            rect.pivot = new Vector2(0, 1);

            var image = go.AddComponent<Image>();
            image.raycastTarget = true;

            var mask = go.AddComponent<Mask>();
            mask.showMaskGraphic = false;

            return go;
        }

        private static GameObject CreateContent(GameObject viewport, RectOffset padding = null)
        {
            var go = new GameObject("Content");
            go.transform.SetParent(viewport.transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchTop();
            rect.pivot = new Vector2(0, 1);

            var layout = go.AddComponent<VerticalLayoutGroup>();
            layout.spacing = 10f;
            layout.padding = padding ?? new RectOffset(0, 8, 0, 0);
            layout.childForceExpandWidth = true;
            layout.childForceExpandHeight = false;
            layout.childControlWidth = true;
            layout.childControlHeight = true;
            layout.childAlignment = TextAnchor.UpperLeft;

            var fit = go.AddComponent<ContentSizeFitter>();
            fit.horizontalFit = ContentSizeFitter.FitMode.Unconstrained;
            fit.verticalFit = ContentSizeFitter.FitMode.PreferredSize;

            return go;
        }

        public AudioMateController c;
        private readonly List<JSONStorableParam> _storables = new List<JSONStorableParam>();

        private static string PrepareLabelStyle(string label)
        {
            //return label.ToUpper();
            return label;
        }

        public UIDynamic CreateSpacer(float height = 20f)
        {
            var ui = Instantiate(c.Manager.configurableSpacerPrefab).GetComponent<UIDynamic>();
            ui.gameObject.transform.SetParent(transform, false);
            LayoutElement layout = ui.GetComponent<LayoutElement>();
            layout.minHeight = 0;
            ui.height = height;
            return ui;
        }

        public GameObject CreateLayoutContainer(Transform parent = null, float scale = 1f, bool vertical = true)
        {
            var go = new GameObject();
            go.transform.SetParent((Object) parent != (Object) null ? parent : transform, false);
            var rect = go.AddComponent<RectTransform>();
            rect.StretchParent();
            if (vertical)
            {
                var layoutGroup = go.AddComponent<VerticalLayoutGroup>();
                layoutGroup.childForceExpandWidth = true;
                layoutGroup.childForceExpandHeight = false;
                layoutGroup.childControlWidth = true;
                layoutGroup.padding = new RectOffset(6, 9, 0, 0);
            }
            else
            {
                var layoutGroup = go.AddComponent<HorizontalLayoutGroup>();
                layoutGroup.childForceExpandWidth = true;
                layoutGroup.childForceExpandHeight = false;
                layoutGroup.childControlHeight = true;
                layoutGroup.padding = new RectOffset(6, 9, 0, 0);
            }

            return go;
        }

        public UIDynamicSlider CreateSlider(JSONStorableFloat jsf, float scale = 1.0f, bool showQuickButtons = false, string valFormat = "F2", bool doNotRegister = false)
        {
            if (!doNotRegister) RegisterStorable(jsf);
            var prefab = c.Manager.configurableSliderPrefab;
            prefab.localScale = new Vector3(scale, scale);
            var ui = Instantiate(c.Manager.configurableSliderPrefab).GetComponent<UIDynamicSlider>();
            ui.gameObject.transform.SetParent(transform, false);
            var uiText = ui.sliderValueTextFromFloat.UIText;
            if ((Object) uiText != (Object) null) uiText.color = Styles.ButtonText;
            var uiInput = ui.sliderValueTextFromFloat.UIInputField;
            uiInput.caretColor = Styles.SliderText;
            uiInput.selectionColor = Styles.SliderBg;
            uiInput.textComponent.color = Styles.SliderText;
            var rectTransform = ui.gameObject.GetComponent<RectTransform>();
            rectTransform.pivot = new Vector2(0, 0.5f);
            var image = rectTransform.GetChild(0).gameObject.GetComponent<Image>();
            image.color = Styles.SliderBg;
            ui.labelText.color = Styles.SliderText;
            ui.Configure(jsf.name, jsf.min, jsf.max, jsf.val, jsf.constrained, valFormat, showQuickButtons, !jsf.constrained);
            jsf.slider = ui.slider;
            return ui;
        }

        public UIDynamicButton CreateButton(string label, int style = Styles.Button, bool addOutline = false)
        {
            var ui = Instantiate(c.Manager.configurableButtonPrefab).GetComponent<UIDynamicButton>();
            ui.buttonColor = Styles.Bg(style);
            ui.textColor = Styles.Text(style);
            ui.buttonText.fontSize = 30;
            if (addOutline)
            {
                var outline = ui.button.gameObject.AddComponent<Outline>().GetComponent<Outline>();
                outline.effectColor = new Color(0f,0f,0f,0.5f);
                outline.effectDistance = new Vector2(2f,2f);
                outline.enabled = true;
            }

            ui.gameObject.transform.SetParent(transform, false);
            ui.label = PrepareLabelStyle(label);
            return ui;
        }

        public HorizontalButtonGroup CreateHorizontalGroup(GameObject go = null)
        {
            return go == null ? gameObject.AddComponent<HorizontalButtonGroup>() : go.AddComponent<HorizontalButtonGroup>();
        }

        public UIDynamicToggle CreateToggle(JSONStorableBool jsb, int style = Styles.Toggle)
        {
            RegisterStorable(jsb);
            var ui = Instantiate(c.Manager.configurableTogglePrefab).GetComponent<UIDynamicToggle>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.backgroundColor = Styles.Bg(style);
            ui.textColor = Styles.Text(style);
            ui.label = jsb.name;
            ui.toggle.graphic.color = Styles.Bg(style);
            jsb.toggle = ui.toggle;
            return ui;
        }

        public UIDynamicTextField CreateTextField(JSONStorableString jss)
        {
            RegisterStorable(jss);
            var ui = Instantiate(c.Manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.UItext.fontSize = 28;
            ui.backgroundColor = Styles.EmbeddedBg;
            ui.textColor = Styles.EmbeddedText;
            ui.backgroundImage.fillMethod = Image.FillMethod.Radial180;
            jss.dynamicText = ui;
            return ui;
        }

        public UIDynamicPopup CreatePopup(JSONStorableStringChooser jsc, bool filterable, float panelHeight = 350f, bool doNotRegister = false)
        {
            if (!doNotRegister) RegisterStorable(jsc);
            Transform prefab;
            #if(VAM_GT_1_20)
            if(filterable)
                prefab = c.manager.configurableFilterablePopupPrefab;
            else
            #endif
                prefab = c.Manager.configurableScrollablePopupPrefab;
            var ui = Instantiate(prefab).GetComponent<UIDynamicPopup>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.labelTextColor = Styles.PopupText;
            ui.popup.backgroundImage.color = Styles.PopupBg;
            ui.label = jsc.name;
            ui.popup.label = jsc.val;
            ui.popupPanelHeight = panelHeight;
            jsc.popup = ui.popup;
            return ui;
        }

        public UIDynamicTextField CreateTextInput(
            JSONStorableString jss,
            bool hasLabel = false,
            InputField.ContentType contentType = InputField.ContentType.Alphanumeric)
        {
            RegisterStorable(jss);

            var go = new GameObject();
            go.transform.SetParent(transform, false);
            {
                var rect = go.AddComponent<RectTransform>();
                rect.pivot = new Vector2(0, 1);

                var layout = go.AddComponent<LayoutElement>();
                layout.preferredHeight = 25f;
                layout.flexibleHeight = 0;
                layout.flexibleWidth = 1f;
            }

            var textfield = Instantiate(c.Manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();
            textfield.gameObject.transform.SetParent(go.transform, false);
            {
                jss.dynamicText = textfield;

                textfield.backgroundColor = Styles.InputBg;
                textfield.textColor = Styles.InputText;
                var input = textfield.gameObject.AddComponent<InputField>();
                textfield.UItext.resizeTextForBestFit = true;
                textfield.UItext.fontSize = 42;
                textfield.UItext.alignment = TextAnchor.LowerLeft;
                textfield.UItext.alignByGeometry = true;
                input.textComponent = textfield.UItext;
                input.contentType = contentType;
                jss.inputField = input;
                input.lineType = InputField.LineType.SingleLine;
                //input.textComponent.alignment = TextAnchor.LowerLeft;
                var rect = textfield.GetComponent<RectTransform>();
                rect.anchorMin = new Vector2(0, 0);
                rect.anchorMax = new Vector2(1, 1);
                rect.pivot = new Vector2(0, 0);
                rect.anchoredPosition = new Vector2(0, 0);
                rect.sizeDelta = new Vector2(0, 25f);
                var layout = textfield.GetComponent<LayoutElement>();
                Destroy(layout);
            }
            if (!hasLabel) return textfield;
            var title = new GameObject();
            title.transform.SetParent(go.transform, false);
            {
                var rect = title.AddComponent<RectTransform>();
                rect.anchorMin = new Vector2(0, 0);
                rect.anchorMax = new Vector2(1, 1);
                rect.pivot = new Vector2(0, 1);
                rect.anchoredPosition = new Vector2(0, 1f);
                rect.sizeDelta = new Vector2(0, 30f);

                var text = title.AddComponent<Text>();
                text.font = textfield.UItext.font;
                text.text = jss.name;
                text.fontSize = 24;
                text.color = Styles.ButtonBg;
            }
            return textfield;
        }

        public UIDynamicTextField CreateSimpleTextInput(
            JSONStorableString jss,
            int fontSize = 42,
            InputField.ContentType contentType = InputField.ContentType.Alphanumeric)
        {
            RegisterStorable(jss);
            var go = new GameObject();


                go.transform.SetParent(transform, false);

                var rect = go.AddComponent<RectTransform>();
                rect.pivot = new Vector2(0, 1);
                var blayout = go.gameObject.AddComponent<LayoutElement>();
                blayout.preferredHeight = fontSize * 1.2f;
                blayout.flexibleHeight = 0;
                blayout.flexibleWidth = 1;


        var textfield = Instantiate(c.Manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();
            textfield.gameObject.transform.SetParent(go.transform, false);

                jss.dynamicText = textfield;

                textfield.backgroundColor = Palette.TransparentLightGray;
                textfield.textColor = Palette.TextAlt;
                var input = textfield.gameObject.AddComponent<InputField>();
                textfield.UItext.resizeTextForBestFit = true;
                textfield.UItext.fontSize = fontSize;
                textfield.UItext.fontStyle = FontStyle.Bold;
                textfield.UItext.alignment = TextAnchor.MiddleLeft;
                textfield.UItext.fontStyle = FontStyle.Bold;

                input.textComponent = textfield.UItext;
                input.contentType = contentType;
                jss.inputField = input;
                input.lineType = InputField.LineType.SingleLine;
                input.textComponent.alignment = TextAnchor.MiddleLeft;
                var rect2 = textfield.GetComponent<RectTransform>();
                rect2.anchorMin = new Vector2(0, 0);
                rect2.anchorMax = new Vector2(1, 1);
                rect2.pivot = new Vector2(0, 0);
                rect2.anchoredPosition = new Vector2(0, 20f);
                rect2.sizeDelta = new Vector2(0, 0);
                /* var layout = textfield.GetComponent<LayoutElement>();
                Destroy(layout); */

            return textfield;
        }

        public Text CreateHeader(string headline, int level = 2, float paddingTop = 40f, float paddingBottom = 10f,
            TextAnchor alignment = TextAnchor.LowerLeft, int flexibleHeight = 1, int flexibleWidth = 1)
        {
            if (paddingTop > 0f) CreateSpacer(paddingTop);
            var go = new GameObject();
            go.transform.SetParent(transform, false);
            var rect = go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1);
            rect.StretchParent();
            var layout = go.AddComponent<LayoutElement>();
            var text = go.AddComponent<Text>();
            text.text = PrepareLabelStyle(headline);
            text.font = BaseFont;
            text.color = Styles.Text();
            text.alignment = alignment;
            if (level > 3) level = 3;
            switch (level)
            {
                case 3:
                    text.fontSize = 28;
                    text.fontStyle = FontStyle.Bold;
                    break;
                case 2:
                    text.fontSize = 34;
                    text.fontStyle = FontStyle.Bold;
                    break;
                default:
                    text.fontSize = 48;
                    text.fontStyle = FontStyle.Bold;
                    break;
            }
            layout.flexibleHeight = flexibleHeight;
            layout.flexibleWidth = flexibleWidth;
            layout.preferredHeight = text.fontSize + 10;
            layout.minHeight = text.fontSize + 10;
            if (paddingBottom > 0f) CreateSpacer(paddingBottom);

            return text;
        }
        public Text CreateText(string textContent, int level = 3, int padding = 0, TextAnchor alignment = TextAnchor.UpperLeft)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);
            var rect = go.AddComponent<RectTransform>();
            /* rect.pivot = new Vector2(0, 1);
            rect.StretchParent(); */
            var layout = go.AddComponent<LayoutElement>();
            var text = rect.gameObject.AddComponent<Text>();
            text.text = PrepareLabelStyle(textContent);
            text.font = BaseFont;
            text.color = Styles.Text();
            text.alignment = alignment;
            if (level > 3) level = 3;
            switch (level)
            {
                case 3:
                    text.fontSize = 24;
                    text.fontStyle = FontStyle.Normal;
                    break;
                case 2:
                    text.fontSize = 32;
                    text.fontStyle = FontStyle.Normal;
                    break;
                default:
                    text.fontSize = 48;
                    text.fontStyle = FontStyle.Normal;
                    break;
            }

            layout.preferredHeight = text.fontSize * (text.text.Length / (text.fontSize * 1.5f)) + padding;
            layout.flexibleHeight = layout.preferredHeight;
            layout.flexibleWidth = 1;
            layout.minHeight = layout.preferredHeight;

            return text;
        }

        public void ClearOverlay()
        {
            Destroy(_currentContainer);
            _currentContainer = null;
        }

        public void CreateModal(string message, OverlaySettings settings , Action[] buttonCallbacks = null, string heading = null)
        {
            if ((Object) _currentContainer != (Object) null) throw new InvalidOperationException("Only one overlay can be shown at a time.");

            var container = new GameObject();
            container.transform.SetParent(transform.parent.parent, false);

            {
                var rect = container.AddComponent<RectTransform>();
                rect.StretchParent();

                var bg = container.AddComponent<Image>();
                bg.color = Styles.OverlayBg;
                bg.preserveAspect = false;
            }

            var modal = new GameObject();
            modal.transform.SetParent(container.transform, false);
            {   var rect = modal.AddComponent<RectTransform>();
                rect.StretchParent();
                rect.sizeDelta = new Vector2(-350, -500);
                var bg = modal.AddComponent<Image>();
                bg.color = Styles.ModalBg;
                bg.preserveAspect = false;
            }

            var modalContent = new GameObject();
            modalContent.transform.SetParent(modal.transform, false);
            {
                var rect = modalContent.AddComponent<RectTransform>();
                rect.StretchParent();
                var layout = modalContent.AddComponent<VerticalLayoutGroup>();
                layout.childControlHeight = true;
                layout.childForceExpandHeight = false;
                layout.childControlWidth = true;
                layout.childForceExpandWidth = false;
                layout.spacing = 10f;
                layout.padding = new RectOffset(50, 50, 50, 50);
                if (heading != null) {
                    var headerText = CreateHeader(heading, 1, 0f, 0f, TextAnchor.MiddleCenter, 0);
                    headerText.transform.SetParent(modalContent.transform, false);
                    headerText.alignment = TextAnchor.MiddleCenter;
                    headerText.color = Styles.ModalText;
                }
                var text = CreateText(message, 2);
                text.transform.SetParent(modalContent.transform, false);
                text.color = Styles.ModalText;
                text.alignment = TextAnchor.MiddleCenter;
            }

            var mainButtons = settings.MainButtons;
            for (var index = 0; index < mainButtons.Count; index++)
            {
                var button = mainButtons[index];
                var btn = CreateButton(button.Label);
                btn.transform.SetParent(modalContent.transform, false);
                btn.buttonImage.fillCenter = false;

                btn.buttonColor = Styles.Text(button.Style);
                btn.textColor = Styles.Text(button.Style);
                var callback = buttonCallbacks?[index] ?? button.Callback;
                btn.button.onClick.AddListener(() =>
                {
                    _currentContainer = null;
                    Destroy(container);
                    callback?.Invoke();
                });
            }

            _currentContainer = container;
        }

         public RectTransform CreateTriggerActionsUI() {
             RectTransform rt = null;
             if (_triggerActionsPrefab != null)
             {
                 rt = (RectTransform)Instantiate(_triggerActionsPrefab);
             } else {
                 Debug.LogError("Attempted to make TriggerActionsUI when prefab was not set");
             }
             return (rt);
         }

         public RectTransform CreateCollectionTriggerActionsUI()
         {
             var rt = CreateTriggerActionsUI();
             var rects = rt.GetComponentsInChildren<RectTransform>();
             foreach (var rectTransform in rects)
             {
                 if (rectTransform.gameObject.name != "Tab2") continue;
                 rectTransform.gameObject.SetActive(false);
                 break;
             }
             // Update labels and texts
             var texts = rt.gameObject.GetComponentsInChildren<Text>();
             foreach (var text in texts)
             {
                 var parentName = text.transform.parent.gameObject.name;
                 switch (parentName)
                 {
                     case "AddDiscreteStartActionButton":
                         text.text = "Add Collection Trigger Action";
                         break;
                     case "Trigger Name Text":
                         text.text = "Collection Trigger Actions";
                         break;
                 }

                 //c.DebugUI(text.gameObject);
             }
             return rt;
         }

         public RectTransform CreateCollectionTriggerActionMiniUI()
         {
             var rt = CreateTriggerActionMiniUI();
             var buttons = rt.gameObject.GetComponentsInChildren<Button>();
             foreach (var button in buttons)
             {
                 var parentName = button.transform.parent.gameObject.name;
                 switch (button.name)
                 {
                     case "Open Button":
                         button.gameObject.SetActive(false);
                         break;
                 }

                 //c.DebugUI(text.gameObject);
             }
             return (rt);
         }

         public RectTransform CreateTriggerActionMiniUI() {
             RectTransform rt = null;
             if (_triggerActionMiniPrefab != null) {
                 rt = (RectTransform)Instantiate(_triggerActionMiniPrefab);
             } else {
                 Debug.LogError("Attempted to make TriggerActionMiniUI when prefab was not set");
             }
             return (rt);
         }

         public RectTransform CreateTriggerActionDiscreteUI() {
             RectTransform rt = null;
             if (_triggerActionDiscretePrefab != null) {
                 rt = (RectTransform)Instantiate(_triggerActionDiscretePrefab);
             } else {
                 Debug.LogError("Attempted to make TriggerActionDiscreteUI when prefab was not set");
             }
             return (rt);
         }

         public RectTransform CreateTriggerActionTransitionUI() {
             RectTransform rt = null;
             if (_triggerActionTransitionPrefab != null) {
                 rt = (RectTransform)Instantiate(_triggerActionTransitionPrefab);
             } else {
                 Debug.LogError("Attempted to make TriggerActionTransitionUI when prefab was not set");
             }
             return (rt);
         }

         public void RemoveTriggerActionUI(RectTransform rt) {
             if (rt != null) {
                 Destroy(rt.gameObject);
             }
         }

         public void OnDestroy()
        {
            var clone = new JSONStorableParam[_storables.Count];
            _storables.CopyTo(clone);
            _storables.Clear();
            foreach (var component in clone)
            {
                if (component == null) continue;
                if (component is JSONStorableStringChooser)
                    RemovePopup((JSONStorableStringChooser)component);
                else if (component is JSONStorableFloat)
                    RemoveSlider((JSONStorableFloat)component);
                else if (component is JSONStorableString)
                    RemoveTextField((JSONStorableString)component);
                else if (component is JSONStorableBool)
                    RemoveToggle((JSONStorableBool)component);
                else
                    SuperController.LogError($"AudioMate: Cannot remove component {component}");
            }
        }

        public void RemovePopup(JSONStorableStringChooser jsc, UIDynamicPopup component = null)
        {
            if (jsc.popup != null) { jsc.popup = null; _storables.Remove(jsc); }
            if (component != null) Destroy(component.gameObject);
        }

        public void RemoveSlider(JSONStorableFloat jsf, UIDynamicSlider component = null)
        {
            if (jsf.slider != null) { jsf.slider = null; _storables.Remove(jsf); }
            if (component != null) Destroy(component.gameObject);
        }

        public void RemoveTextField(JSONStorableString jss, UIDynamicTextField component = null)
        {
            if (jss.dynamicText != null) { jss.dynamicText = null; _storables.Remove(jss); }
            if (component != null) Destroy(component.gameObject);
        }

        public void RemoveToggle(JSONStorableBool jsb, UIDynamicToggle component = null)
        {
            if (jsb.toggle != null) { jsb.toggle = null; _storables.Remove(jsb); }
            if (component != null) Destroy(component.gameObject);
        }

        private T RegisterStorable<T>(T v)
            where T : JSONStorableParam
        {
            _storables.Add(v);
            ValidateStorableFreeToBind(v);
            return v;
        }

        private void ValidateStorableFreeToBind(JSONStorableParam v)
        {
            if (v is JSONStorableStringChooser)
            {
                if (((JSONStorableStringChooser)v).popup != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
            else if (v is JSONStorableFloat)
            {
                if (((JSONStorableFloat)v).slider != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
            else if (v is JSONStorableString)
            {
                if (((JSONStorableString)v).inputField != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
            else if (v is JSONStorableBool)
            {
                if (((JSONStorableBool)v).toggle != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
        }
    }
}
